from dicetables.baseevents import IntegerEvents


class ProtoDie(IntegerEvents):
    """
    base object for any kind of die.


    :all Die objects need: get_size(), get_weight(), weight_info(), multiply_str(), get_dict()
    :get_dict(): must return {int: int >0}

    """
    def __init__(self):
        super(ProtoDie, self).__init__()

    def get_size(self):
        raise NotImplementedError

    def get_weight(self):
        raise NotImplementedError

    def get_dict(self):
        raise NotImplementedError

    def weight_info(self):
        """return detailed info of how the die is weighted"""
        raise NotImplementedError

    def multiply_str(self, number):
        """return a string that is the die string multiplied by a number. i.e.,
        D6+1 times 3 is '3D6+3' """
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __hash__(self):
        return hash('hash of {!r}, {}, {}, {}'.format(self,
                                                      self.get_size(),
                                                      self.get_weight(),
                                                      self.get_dict()))

    def __lt__(self, other):
        return (
            (self.get_size(), self.get_weight(), sorted(self.get_dict().items()), repr(self)) <
            (other.get_size(), other.get_weight(), sorted(other.get_dict().items()), repr(other))
        )

    def __eq__(self, other):
        if not isinstance(other, ProtoDie):
            return False
        return (
            (self.get_size(), self.get_weight(), sorted(self.get_dict().items()), repr(self)) ==
            (other.get_size(), other.get_weight(), sorted(other.get_dict().items()), repr(other))
        )

    def __ne__(self, other):
        return not self == other

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        return not self <= other

    def __ge__(self, other):
        return self == other or self > other


class Die(ProtoDie):
    """
    stores and returns info for a basic Die.
    Die(4) rolls 1, 2, 3, 4 with equal weight
    """
    def __init__(self, die_size):
        """

        :param die_size: int > 0
        """
        self._die_size = die_size
        super(Die, self).__init__()

    def get_size(self):
        return self._die_size

    def get_weight(self):
        return 0

    def get_dict(self):
        return dict.fromkeys(range(1, self._die_size + 1), 1)

    def weight_info(self):
        return '{}\n    No weights'.format(self)

    def multiply_str(self, number):
        return '{}{}'.format(number, self)

    def __str__(self):
        return 'D{}'.format(self._die_size)

    def __repr__(self):
        return 'Die({})'.format(self._die_size)


class ModDie(Die):
    """
    stores and returns info for a Die with a modifier
    that changes the values of the rolls.
    ModDie(4, -1) rolls 0, 1, 2, 3 with equal weight
    """
    def __init__(self, die_size, modifier):
        """

        :param die_size: int >0
        :param modifier: int
        """
        self._mod = modifier
        super(ModDie, self).__init__(die_size)

    def get_modifier(self):
        return self._mod

    def get_dict(self):
        return dict.fromkeys(range(1 + self._mod, self.get_size() + 1 + self._mod), 1)

    def multiply_str(self, number):
        return '{}D{}{:+}'.format(number, self.get_size(), number * self._mod)

    def __str__(self):
        return 'D{}{:+}'.format(self.get_size(), self._mod)

    def __repr__(self):
        return 'ModDie({}, {})'.format(self.get_size(), self._mod)


class WeightedDie(ProtoDie):
    """
    stores and returns info for die with different chances for different rolls.
    WeightedDie({1:1, 2:5}) rolls 1 once for every five times that 2 is rolled.
    """
    def __init__(self, dictionary_input):
        """

        :param dictionary_input: {roll: weight} roll: int>1, weight: int>=0\n
            the sum of all weights >0
        """

        self._raw_dic = self._create_raw_dict(dictionary_input)
        super(WeightedDie, self).__init__()

    def _create_raw_dict(self, dictionary):
        self._raise_value_error_for_rolls_less_than_one(dictionary)
        output = {}
        for key in range(1, max(dictionary.keys()) + 1):
            output[key] = dictionary.get(key, 0)
        return output

    @staticmethod
    def _raise_value_error_for_rolls_less_than_one(dictionary):
        if any(roll < 1 for roll in dictionary):
            raise ValueError('rolls may not be less than 1. use ModWeightedDie')

    def get_raw_dict(self):
        return self._raw_dic.copy()

    def get_size(self):
        return max(self._raw_dic.keys())

    def get_weight(self):
        return sum(self._raw_dic.values())

    def get_dict(self):
        return {key: value for key, value in self._raw_dic.items() if value}

    def weight_info(self):
        max_roll_str_len = len(str(self.get_size()))
        out = str(self) + '\n'
        for roll, weight in sorted(self.get_raw_dict().items()):
            out += '    a roll of {:>{}} has a weight of {}\n'.format(roll, max_roll_str_len, weight)
        return out.rstrip('\n')

    def multiply_str(self, number):
        return '{}{}'.format(number, self)

    def __str__(self):
        return 'D{}  W:{}'.format(self.get_size(), self.get_weight())

    def __repr__(self):
        return 'WeightedDie({})'.format(self._raw_dic)


class ModWeightedDie(WeightedDie):
    """
    stores and returns info for die with different chances for different rolls.
    The modifier changes all die rolls
    WeightedDie({1:1, 2:5}, -1) rolls 0 once for every five times that 1 is rolled.
    """

    def __init__(self, dictionary_input, modifier):
        """

        :param dictionary_input: {roll: weight} roll: int, weight: int>=0\n
            sum of all weights >0
        :param modifier: int
        """
        self._mod = modifier
        super(ModWeightedDie, self).__init__(dictionary_input)

    def get_modifier(self):
        return self._mod

    def get_dict(self):
        return {roll + self._mod: weight for roll, weight in self.get_raw_dict().items() if weight}

    def multiply_str(self, number):
        return '{}D{}{:+}  W:{}'.format(number, self.get_size(),
                                        number * self._mod, self.get_weight())

    def __str__(self):
        return 'D{}{:+}  W:{}'.format(self.get_size(), self._mod, self.get_weight())

    def __repr__(self):
        return 'ModWeightedDie({}, {})'.format(self.get_raw_dict(), self._mod)


class StrongDie(ProtoDie):
    """
    stores and returns info for a stronger version of another die.
    The multiplier multiplies all die rolls of original Die.
    StrongDie(ModDie(3, -1), 2) rolls (1-1)*2, (2-1)*2, (3-1)*2 with equal weight.
    """

    def __init__(self, input_die, multiplier):
        """

        :param input_die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or subclass of ProtoDie
        :param multiplier: int
        """
        self._multiplier = multiplier
        self._original = input_die
        super(StrongDie, self).__init__()

    def get_size(self):
        return self._original.get_size()

    def get_weight(self):
        return self._original.get_weight()

    def get_multiplier(self):
        return self._multiplier

    def get_input_die(self):
        """returns an instance of the original die"""
        return self._original

    def get_dict(self):
        return {roll * self._multiplier: weight for roll, weight in self._original.get_dict().items()}

    def weight_info(self):
        return self._original.weight_info().replace(str(self._original), str(self))

    def multiply_str(self, number):
        return '({})X({})'.format(self._original.multiply_str(number), self._multiplier)

    def __str__(self):
        return '({})X({})'.format(self._original, self._multiplier)

    def __repr__(self):
        return 'StrongDie({!r}, {})'.format(self._original, self._multiplier)



""" this module contains the class LongIntTable and longintmath that the table
needs to deal with it's BFN"""
from __future__ import absolute_import

from sys import version_info

from dicetables.factory.eventsfactory import EventsFactory
from dicetables.tools.dictcombiner import DictCombiner
from dicetables.tools.eventerrors import InvalidEventsError

if version_info[0] < 3:
    from dicetables.tools.py2funcs import is_int
else:
    from dicetables.tools.py3funcs import is_int


class EventsVerifier(object):
    def __init__(self):
        self._type_str = 'ints'
        if version_info[0] < 3:
            self._type_str += ' or longs'

    def verify_get_dict(self, events_dict):
        """

        :param events_dict: IntegerEvents.get_dict()
        :raises: InvalidEventsError
        """
        if not events_dict:
            raise InvalidEventsError('events may not be empty. a good alternative is the identity - {0: 1}.')
        if not self.is_all_ints(events_dict.keys()) or not self.is_all_ints(events_dict.values()):
            raise InvalidEventsError('all values must be {}'.format(self._type_str))
        if any(occurrence <= 0 for occurrence in events_dict.values()):
            raise InvalidEventsError('no negative or zero occurrences in Events.get_dict()')

    @staticmethod
    def is_all_ints(iterable):
        return all(is_int(value) for value in iterable)


class IntegerEvents(object):
    def __init__(self):
        super(IntegerEvents, self).__init__()
        EventsVerifier().verify_get_dict(self.get_dict())

    def get_dict(self):
        """

        :return: {event: occurrences}
        """
        message = ('get_dict() must return a dictionary\n' +
                   '{event: occurrences, ...} event=int, occurrence=int>0.')
        raise NotImplementedError(message)


def scrub_zeroes(dictionary):
    return {key: val for key, val in dictionary.items() if val}


class AdditiveEvents(IntegerEvents):

    def __init__(self, events_dict):
        """

        :param events_dict: {event: occurrences}\n
            event=int. occurrences=int >=0
            total occurrences > 0
        """
        self._table = scrub_zeroes(events_dict)
        super(AdditiveEvents, self).__init__()
        EventsFactory.check(self.__class__)

    @classmethod
    def new(cls):
        return EventsFactory.new(cls)

    def get_dict(self):
        return self._table.copy()

    def __str__(self):
        min_event = min(self._table.keys())
        max_event = max(self._table.keys())
        return 'table from {} to {}'.format(min_event, max_event)

    def combine(self, events, times=1):
        dictionary = self._create_constructor_dict(events, times, method_str='combine')
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_flattened_list(self, events, times=1):
        """

        :WARNING - UNSAFE METHOD: len(flattened_list) = total occurrences of events.
            if this list is too big, it will raise MemoryError or OverflowError
        """
        dictionary = self._create_constructor_dict(events, times, method_str='combine_by_flattened_list')
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_dictionary(self, events, times=1):
        dictionary = self._create_constructor_dict(events, times, method_str='combine_by_dictionary')
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_indexed_values(self, events, times=1):
        dictionary = self._create_constructor_dict(events, times, method_str='combine_by_indexed_values')
        return EventsFactory.from_dictionary(self, dictionary)

    def remove(self, events, times=1):
        """

        :WARNING - UNSAFE METHOD: There is no record of what you added to an AdditiveEvents.
            If you remove what you haven't added, no error will be raised, but you will have bugs.
        """
        dictionary = self._create_constructor_dict(events, times, method_str='remove')
        return EventsFactory.from_dictionary(self, dictionary)

    def _create_constructor_dict(self, events, times, method_str):
        combiner = DictCombiner(self.get_dict())
        methods = {'combine': combiner.combine_by_fastest,
                   'combine_by_flattened_list': combiner.combine_by_flattened_list,
                   'combine_by_dictionary': combiner.combine_by_dictionary,
                   'combine_by_indexed_values': combiner.combine_by_indexed_values,
                   'remove': combiner.remove_by_tuple_list}
        new_combiner = methods[method_str](events.get_dict(), times)
        return new_combiner.get_dict()



"""functions for getting useful info from tables, ie. plot pts or stats"""
from __future__ import absolute_import

from decimal import Decimal
from math import log10

from dicetables.tools.numberforamtter import NumberFormatter
from dicetables.tools.listtostring import get_string_from_list_of_ints


def safe_true_div(numerator, denominator):
    """floating point division for any sized number"""
    ans = Decimal(numerator) / Decimal(denominator)
    return _convert_decimal_to_float_or_int(ans)


def _convert_decimal_to_float_or_int(num):
    answer_as_float = float(num)
    if answer_as_float == float('inf') or answer_as_float == float('-inf'):
        return int(num)
    else:
        return answer_as_float


class EventsInformation(object):
    def __init__(self, events):
        self._dict = events.get_dict()

    def get_items(self):
        return self._dict.items()

    def events_keys(self):
        return sorted(self._dict.keys())

    def events_range(self):
        all_keys = self.events_keys()
        return all_keys[0], all_keys[-1]

    def total_occurrences(self):
        return sum(self._dict.values())

    def all_events(self):
        return sorted(self._dict.items())

    def all_events_include_zeroes(self):
        start, stop = self.events_range()
        return self.get_range_of_events(start, stop + 1)

    def biggest_event(self):
        """

        :return: (event, occurrences) for first event with highest occurrences
        """
        highest_occurrences = max(self._dict.values())
        for event, occurrences in sorted(self._dict.items()):
            if occurrences == highest_occurrences:
                return event, highest_occurrences

    def biggest_events_all(self):
        highest_occurrences = max(self._dict.values())
        output = []
        for event, occurrences in sorted(self._dict.items()):
            if occurrences == highest_occurrences:
                output.append((event, occurrences))
        return sorted(output)

    def get_event(self, event):
        return event, self._dict.get(event, 0)

    def get_range_of_events(self, start, stop_before):
        return [self.get_event(event) for event in range(start, stop_before)]


class EventsCalculations(object):

    def __init__(self, events, include_zeroes=True):
        self._info = EventsInformation(events)
        self._include_zeroes = include_zeroes

    @property
    def include_zeroes(self):
        return self._include_zeroes

    @property
    def info(self):
        return self._info

    def mean(self):
        numerator = sum((value * freq) for value, freq in self._info.get_items())
        denominator = self._info.total_occurrences()
        return safe_true_div(numerator, denominator)

    def stddev(self, decimal_place=4):
        avg = self.mean()
        factor_to_truncate_digits = self._get_truncation_factor(decimal_place)
        truncated_deviations = 0
        for event_value, occurrences in self._info.get_items():
            truncated_deviations += (occurrences // factor_to_truncate_digits) * (avg - event_value) ** 2.
        truncated_total_occurrences = self._info.total_occurrences() // factor_to_truncate_digits
        return round((truncated_deviations / truncated_total_occurrences) ** 0.5, decimal_place)

    def _get_truncation_factor(self, decimal_place):
        extra_digits = 5
        largest_exponent = int(log10(self._info.biggest_event()[1]))
        required_exp_for_accuracy = 2 * (extra_digits + decimal_place)
        if largest_exponent < required_exp_for_accuracy:
            factor_to_truncate_digits = 1
        else:
            factor_to_truncate_digits = 10 ** (largest_exponent - required_exp_for_accuracy)
        return factor_to_truncate_digits

    def percentage_points(self):
        return self._percentage_points_by_method('fast')

    def percentage_points_exact(self):
        return self._percentage_points_by_method('exact')

    def percentage_axes(self):
        return list(zip(*self.percentage_points()))

    def percentage_axes_exact(self):
        return list(zip(*self.percentage_points_exact()))

    def _percentage_points_by_method(self, method_str):
        methods = {'fast': get_fast_pct_number, 'exact': get_exact_pct_number}
        pct_method = methods[method_str]
        total_values = self._info.total_occurrences()
        return [(event, pct_method(occurrence, total_values)) for event, occurrence in self._get_data_set()]

    def _get_data_set(self):
        if self._include_zeroes:
            return self._info.all_events_include_zeroes()
        else:
            return self._info.all_events()

    def full_table_string(self):
        formatter = NumberFormatter()
        right_just = self._get_right_just()
        out_str = ''
        for value, frequency in self._get_data_set():
            out_str += '{:>{}}: {}\n'.format(value, right_just, formatter.format(frequency))
        return out_str

    def _get_right_just(self):
        min_event, max_event = self._info.events_range()
        value_right_just = max(len(str(min_event)), len(str(max_event)))
        return value_right_just

    def stats_strings(self, query_values):
        """

        :return: (query values, query occurrences, total occurrences, inverse chance, pct chance)
        """
        formatter = NumberFormatter()

        total_occurrences = self._info.total_occurrences()
        query_values_occurrences = self._get_query_values_occurrences(query_values)

        inverse_chance_str, pct_str = _calculate_chance_and_pct(query_values_occurrences, total_occurrences)

        return (get_string_from_list_of_ints(query_values),
                formatter.format(query_values_occurrences),
                formatter.format(total_occurrences),
                formatter.format(inverse_chance_str),
                formatter.format(pct_str))

    def _get_query_values_occurrences(self, query_values):
        combinations_of_values = 0
        no_copies = set(query_values)
        for value in no_copies:
            combinations_of_values += self._info.get_event(value)[1]
        return combinations_of_values


def _calculate_chance_and_pct(query_values_occurrences, total_combinations):
    if not query_values_occurrences:
        return Decimal('inf'), Decimal('0')
    inverse_chance = Decimal(total_combinations) / Decimal(query_values_occurrences)
    pct = Decimal(100.0) / inverse_chance
    return inverse_chance, pct


def get_fast_pct_number(number, total_values):
    factor = 10 ** 50
    will_not_overflow = (number * factor) // total_values
    return will_not_overflow * 100. / float(factor)


def get_exact_pct_number(number, total_values):
    return safe_true_div(100 * number, total_values)


# wrappers functions and deprecated functions

def events_range(events):
    return EventsInformation(events).events_range()


def mean(events):
    return EventsCalculations(events).mean()


def stddev(events, decimal_place=4):
    return EventsCalculations(events).stddev(decimal_place=decimal_place)


def percentage_points(events, include_zeroes=True):
    return EventsCalculations(events, include_zeroes).percentage_points()


def percentage_axes(events, include_zeroes=True):
    return EventsCalculations(events, include_zeroes).percentage_axes()


def stats(events, query_values):
    return EventsCalculations(events).stats_strings(query_values)


def full_table_string(events, include_zeroes=True):
    return EventsCalculations(events, include_zeroes).full_table_string()




from dicetables.factory.errorhandler import EventsFactoryErrorHandler
from dicetables.factory.warninghandler import EventsFactoryWarningHandler
from dicetables.factory.factorytools import StaticDict, Getter
from dicetables.tools.eventerrors import InvalidEventsError, DiceRecordError


class LoaderError(AttributeError):
    pass


class Loader(object):
    def __init__(self, factory):
        self.factory = factory

    def load(self, new_class):
        if 'factory_keys' in new_class.__dict__:
            factory_keys = new_class.factory_keys
        else:
            raise LoaderError

        new_keys = ()
        if 'new_keys' in new_class.__dict__:
            new_keys = new_class.new_keys

        for key in new_keys:
            self.factory.add_getter(*key)
        self.factory.add_class(new_class, factory_keys)


class EventsFactory(object):

    __default_getters = {'dictionary': Getter('get_dict', {0: 1}),
                         'dice': Getter('dice_data', {}),
                         'calc_bool': Getter('calc_includes_zeroes', True, is_property=True)}

    __default_class_args = {'AdditiveEvents': ('dictionary',),
                            'DiceTable': ('dictionary', 'dice'),
                            'DetailedDiceTable': ('dictionary', 'dice', 'calc_bool')}

    _getters = StaticDict(__default_getters)
    _class_args = StaticDict(__default_class_args)

    @classmethod
    def reset(cls):
        cls._getters = StaticDict(cls.__default_getters)
        cls._class_args = StaticDict(cls.__default_class_args)

    @classmethod
    def has_class(cls, events_class):
        return events_class.__name__ in cls._class_args.keys()

    @classmethod
    def has_getter(cls, getter_key):
        return getter_key in cls._getters.keys()

    @classmethod
    def get_class_params(cls, events_class):
        return cls._class_args.get(events_class.__name__)

    @classmethod
    def get_getter_string(cls, getter_key):
        return cls._getters.get(getter_key).__str__()

    @classmethod
    def get_keys(cls):
        return sorted(cls._class_args.keys()), sorted(cls._getters.keys())

    @classmethod
    def add_class(cls, events_class, getter_key_words):
        cls._check_against_factory_classes(events_class, getter_key_words)
        cls._check_for_missing_getters(events_class, getter_key_words)
        cls._class_args = cls._class_args.set(events_class.__name__, getter_key_words)

    @classmethod
    def _check_against_factory_classes(cls, events_class, class_args_tuple):
        if cls.has_class(events_class) and cls.get_class_params(events_class) != class_args_tuple:
            EventsFactoryErrorHandler(cls).raise_error('CLASS OVERWRITE', events_class, class_args_tuple)

    @classmethod
    def _check_for_missing_getters(cls, events_class, class_args_tuple):
        for getter_key in class_args_tuple:
            if not cls.has_getter(getter_key):
                EventsFactoryErrorHandler(cls).raise_error('MISSING GETTER', events_class, getter_key)

    @classmethod
    def add_getter(cls, getter_key, getter_method, empty_value, type_str='method'):
        is_property = False
        if type_str == 'property':
            is_property = True
        new_getter = Getter(getter_method, empty_value, is_property)
        cls._check_against_factory_getters(getter_key, new_getter)
        cls._getters = cls._getters.set(getter_key, new_getter)

    @classmethod
    def _check_against_factory_getters(cls, getter_key, new_getter):
        if cls.has_getter(getter_key) and cls._getters.get(getter_key) != new_getter:
            EventsFactoryErrorHandler(cls).raise_error('GETTER OVERWRITE', getter_key, new_getter)

    @classmethod
    def check(cls, events_class):
        if not cls.has_class(events_class):
            try:
                Loader(cls).load(events_class)
            except LoaderError:
                EventsFactoryWarningHandler(cls).raise_warning('CHECK', events_class)

    @classmethod
    def from_dictionary(cls, events, dictionary):
        passed_in_values = {'dictionary': dictionary}
        return cls._construct_from(events, passed_in_values)

    @classmethod
    def from_dictionary_and_dice(cls, events, dictionary, dice):
        passed_in_values = {'dictionary': dictionary, 'dice': dice}
        return cls._construct_from(events, passed_in_values)

    @classmethod
    def from_params(cls, events, name_value_param_dict):
        return cls._construct_from(events, name_value_param_dict)

    @classmethod
    def _construct_from(cls, events, passed_in_values):
        constructor_class = events.__class__
        factory_class = cls._get_nearest_factory_class(constructor_class)
        args = cls._get_args(events, factory_class, passed_in_values)
        return cls._construct(constructor_class, factory_class, args)

    @classmethod
    def new(cls, events_class):
        factory_class = cls._get_nearest_factory_class(events_class)
        args = cls._get_default_args(factory_class)
        return cls._construct(events_class, factory_class, args)

    @classmethod
    def _get_nearest_factory_class(cls, events_class):
        if not cls.has_class(events_class):
            try:
                Loader(cls).load(events_class)
            except LoaderError:
                return cls._walk_the_mro(events_class)
        return events_class

    @classmethod
    def _walk_the_mro(cls, failed_class):
        for parent_class in failed_class.mro():
            if cls.has_class(parent_class):
                EventsFactoryWarningHandler(cls).raise_warning('CONSTRUCT', failed_class, parent_class)
                return parent_class
        EventsFactoryErrorHandler(cls).raise_error('WTF', failed_class)
        return object

    @classmethod
    def _get_default_args(cls, in_factory):
        args = []
        for getter_key in cls.get_class_params(in_factory):
            getter = cls._getters.get(getter_key)
            args.append(getter.get_default())
        return args

    @classmethod
    def _get_args(cls, events, factory_class, passed_in_values):
        new_args = []
        for getter_key in cls.get_class_params(factory_class):
            if getter_key in passed_in_values.keys():
                arg_value = passed_in_values[getter_key]
            else:
                getter = cls._getters.get(getter_key)
                arg_value = getter.get_from(events)
            new_args.append(arg_value)
        return new_args

    @classmethod
    def _construct(cls, original_class, factory_class, args):
        try:
            return original_class(*args)
        except (TypeError, AttributeError, DiceRecordError, InvalidEventsError):
            if original_class == factory_class:
                EventsFactoryErrorHandler(cls).raise_error('SIGNATURES DIFFERENT', original_class)
            return factory_class(*args)



"""These are tested in dicetables.baseevents and dicetables.dicetable . They are here so that modules can
access them without needing to import the above mentioned modules."""


class InvalidEventsError(ValueError):
    def __init__(self, message='', *args, **kwargs):
        super(InvalidEventsError, self).__init__(message, *args, **kwargs)


class DiceRecordError(ValueError):
    def __init__(self, message='', *args, **kwargs):
        super(DiceRecordError, self).__init__(message, *args, **kwargs)



"""for all things dicey."""
from __future__ import absolute_import
from sys import version_info

from dicetables.baseevents import AdditiveEvents
from dicetables.eventsinfo import EventsCalculations
from dicetables.factory.eventsfactory import EventsFactory
from dicetables.dieevents import ProtoDie
from dicetables.tools.eventerrors import DiceRecordError

if version_info[0] < 3:
    from dicetables.tools.py2funcs import is_int
else:
    from dicetables.tools.py3funcs import is_int


class RecordVerifier(object):
    @staticmethod
    def check_types(die_input):
        if any(not isinstance(die, ProtoDie) or not is_int(num) for die, num in die_input.items()):
            raise DiceRecordError('input must be {ProtoDie: int, ...}')

    @staticmethod
    def check_negative(die_input):
        for key, val in die_input.items():
            if val < 0:
                raise DiceRecordError('Tried to create a DiceRecord with a negative value at {!r}: {}'.format(key, val))

    @staticmethod
    def check_number(num):
        if num < 0:
            raise DiceRecordError('Tried to add_die or remove_die with a negative number.')


def scrub_zeroes(input_dict):
    return {key: val for key, val in input_dict.items() if val}


class DiceRecord(object):
    def __init__(self, dice_number_dict):
        RecordVerifier.check_types(dice_number_dict)
        RecordVerifier.check_negative(dice_number_dict)
        self._record = scrub_zeroes(dice_number_dict)

    def get_dict(self):
        return self._record.copy()

    def get_number(self, query_die):
        return self._record.get(query_die, 0)

    def add_die(self, die, times):
        RecordVerifier.check_number(times)
        new = self._record.copy()
        new[die] = times + new.get(die, 0)
        return DiceRecord(new)

    def remove_die(self, die, times):
        RecordVerifier.check_number(times)
        new = self._record.copy()
        new[die] = new.get(die, 0) - times
        return DiceRecord(new)


class DiceTable(AdditiveEvents):
    def __init__(self, events_dict, dice_number_dict):
        self._record = DiceRecord(dice_number_dict)
        super(DiceTable, self).__init__(events_dict)

    def dice_data(self):
        return self._record.get_dict()

    def get_list(self):
        """

        :return: sorted copy of dice list: [(die, number of dice), ...]
        """
        return sorted(self._record.get_dict().items())

    def number_of_dice(self, query_die):
        return self._record.get_number(query_die)

    def weights_info(self):
        """

        :return: str: complete info for all dice
        """
        output = ''
        for die, number in self.get_list():
            output += format_die_info(die, number)
        return output.rstrip('\n')

    def __str__(self):
        str_list = [die.multiply_str(number) for die, number in self.get_list()]
        return '\n'.join(str_list)

    def add_die(self, die, times=1):
        """

        :param times: int>= 0
        :param die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or new ProtoDie subclass
        """
        dice_iterable = self._create_constructor_dice(die, times, method_str='add_die')
        dictionary = self._create_constructor_dict(die, times, method_str='combine')
        return EventsFactory.from_dictionary_and_dice(self, dictionary, dice_iterable)

    def remove_die(self, die, times=1):
        """

        :param times: 0 <= int <= number of "die" in table
        :param die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or new ProtoDie subclass
        """
        dice_iterable = self._create_constructor_dice(die, times, method_str='remove_die')
        dictionary = self._create_constructor_dict(die, times, method_str='remove')
        return EventsFactory.from_dictionary_and_dice(self, dictionary, dice_iterable)

    def _create_constructor_dice(self, die, times, method_str):
        methods = {'add_die': self._record.add_die,
                   'remove_die': self._record.remove_die}
        new_record = methods[method_str](die, times)
        return new_record.get_dict()


def format_die_info(die, number):
    weight_info = die.weight_info()
    adjusted_info = weight_info.replace(str(die), die.multiply_str(number)) + '\n\n'
    return adjusted_info


class DetailedDiceTable(DiceTable):
    def __init__(self, events_dict, dice_number_dict, calc_includes_zeroes=True):
        super(DetailedDiceTable, self).__init__(events_dict, dice_number_dict)
        self._calc = EventsCalculations(self, calc_includes_zeroes)

    @property
    def info(self):
        return self._calc.info

    @property
    def calc(self):
        return self._calc

    @property
    def calc_includes_zeroes(self):
        return self._calc.include_zeroes

    def switch_boolean(self):
        return EventsFactory.from_params(self, {'calc_bool': not self.calc_includes_zeroes})




from __future__ import absolute_import


from dicetables.tools.eventerrors import DiceRecordError, InvalidEventsError
from dicetables.dieevents import Die, ModDie, WeightedDie, ModWeightedDie, StrongDie, ProtoDie
from dicetables.dicetable import DiceTable, DetailedDiceTable
from dicetables.baseevents import AdditiveEvents
from dicetables.eventsinfo import (EventsCalculations, EventsInformation,
                                   events_range, mean, stddev, percentage_points, percentage_axes, stats,
                                   full_table_string)



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import
from sys import version_info
import unittest
from dicetables.baseevents import AdditiveEvents, InvalidEventsError, EventsVerifier, scrub_zeroes


class TestBaseEvents(unittest.TestCase):
    def setUp(self):
        self.checker = EventsVerifier()
        self.types_error = 'all values must be ints'
        if version_info[0] < 3:
            self.types_error += ' or longs'

    def tearDown(self):
        del self.checker
        del self.types_error

    def assert_my_regex(self, error_type, regex, func, *args):
        with self.assertRaises(error_type) as cm:
            func(*args)
        error_msg = cm.exception.args[0]
        self.assertEqual(error_msg, regex)

    def test_assert_my_regex(self):
        self.assert_my_regex(ValueError, "invalid literal for int() with base 10: 'a'", int, 'a')

    def test_InvalidEventsError_empty(self):
        error = InvalidEventsError()
        self.assertEqual(str(error), '')
        self.assertEqual(error.args[0], '')

    def test_InvalidEventsError_non_empty(self):
        error = InvalidEventsError('message')
        self.assertEqual(str(error), 'message')
        self.assertEqual(error.args[0], 'message')

    def test_EventsVerifier_verify_get_dict_pass(self):
        self.assertIsNone(self.checker.verify_get_dict({1: 1}))

    def test_EventsVerifier_verify_get_dict_empty(self):
        self.assert_my_regex(InvalidEventsError,
                             'events may not be empty. a good alternative is the identity - {0: 1}.',
                             self.checker.verify_get_dict, {})

    def test_EventsVerifier_verify_get_dict_zero_occurrences(self):
        self.assert_my_regex(InvalidEventsError,
                             'no negative or zero occurrences in Events.get_dict()',
                             self.checker.verify_get_dict, {1: 0, 2: 0})

    def test_EventsVerifier_verify_get_dict_negative_occurrences(self):
        self.assert_my_regex(InvalidEventsError, 'no negative or zero occurrences in Events.get_dict()',
                             self.checker.verify_get_dict, {1: -1})

    def test_EventsVerifier_verify_get_dict_non_int_occurrences(self):
        self.assert_my_regex(InvalidEventsError, self.types_error,
                             self.checker.verify_get_dict, {1: 1.0})

    def test_EventsVerifier_verify_get_dict_non_int_event(self):
        self.assert_my_regex(InvalidEventsError, self.types_error,
                             self.checker.verify_get_dict, {1.0: 1})

    def test_EventsVerifier_is_all_ints_pass(self):
        self.assertTrue(self.checker.is_all_ints([10 ** value for value in range(500)]))

    def test_EventsVerifier_all_ints_fail(self):
        self.assertFalse(self.checker.is_all_ints([1.0, 1, 1, 1, 1, 1]))

    def test_EventsVerifier_does_not_work_if_does_not_follow_minimum_requirements(self):
        self.assertRaises(AttributeError, self.checker.verify_get_dict, 'a')
        self.assertRaises(AttributeError, self.checker.verify_get_dict, [1, 2, 3])
        self.assertRaises(AttributeError, self.checker.verify_get_dict, [(1, 2, 3), (4, 5, 6)])
        self.assertRaises(InvalidEventsError, self.checker.verify_get_dict, {'a': 'b'})

    def test_scrub_zeroes_no_zeroes_dict(self):
        self.assertEqual(scrub_zeroes({1: 2, 3: 4}), {1: 2, 3: 4})

    def test_scrub_zeroes_zeroes_in_dict(self):
        self.assertEqual(scrub_zeroes({1: 2, 3: 0, 4: 1, 5: 0}), {1: 2, 4: 1})

    def test_AdditiveEvents__class_method__new(self):
        self.assertEqual(AdditiveEvents.new().get_dict(), {0: 1})

    def test_AdditiveEvents_init_zero_occurrences_dict_raises_error(self):
        self.assertRaises(InvalidEventsError, AdditiveEvents, {1: 0, 2: 0})

    def test_AdditiveEvents_init_empty_dict_raises_error(self):
        self.assertRaises(InvalidEventsError, AdditiveEvents, {})

    def test_AdditiveEvents_init_bad_dict_raises_error(self):
        self.assertRaises(InvalidEventsError, AdditiveEvents, {1: 1.0})
        self.assertRaises(InvalidEventsError, AdditiveEvents, {1.0: 1})
        self.assertRaises(InvalidEventsError, AdditiveEvents, {1: -1})

    def test_AdditiveEvents_get_dict(self):
        self.assertEqual(AdditiveEvents({1: 2, 3: 4}).get_dict(), {1: 2, 3: 4})

    def test_AdditiveEvents_get_dict_removes_zeroes(self):
        self.assertEqual(AdditiveEvents({1: 2, 3: 0, 4: 1, 5: 0}).get_dict(), {1: 2, 4: 1})

    def test_AdditiveEvents_get_dict_will_not_mutate_original(self):
        events = AdditiveEvents({1: 2, 3: 4})
        dictionary = events.get_dict()
        dictionary[1] = 100
        self.assertEqual(AdditiveEvents({1: 2, 3: 4}).get_dict(), {1: 2, 3: 4})

    def test_AdditiveEvents_input_dict_mutation_has_no_effect(self):
        input_dict = {1: 1}
        events = AdditiveEvents(input_dict)
        input_dict[1] = 'banana'
        self.assertEqual(events.get_dict(), {1: 1})

    def test_AdditiveEvents_string_returns_min_to_max(self):
        table = AdditiveEvents({-1: 1, 2: 1, 5: 1})
        self.assertEqual(str(table), 'table from -1 to 5')

    def test_AdditiveEvents_string_is_in_order_and_ignores_high_zero_values(self):
        table = AdditiveEvents({2: 0, 1: 1, -1: 1, -2: 0})
        self.assertEqual(str(table), 'table from -1 to 1')

    def test_AdditiveEvents_combine_negative_times_does_nothing(self):
        new = AdditiveEvents.new().combine(AdditiveEvents({1: 1}), -1)
        self.assertEqual(new.get_dict(), {0: 1})

    def test_AdditiveEvents_combine_by_dictionary(self):
        to_combine = AdditiveEvents({1: 2, 2: 2})
        new = AdditiveEvents.new().combine_by_dictionary(to_combine, 1)
        self.assertEqual(new.get_dict(), to_combine.get_dict())

    def test_AdditiveEvents_combine_by_flattened_list(self):
        to_combine = AdditiveEvents({1: 1, 2: 2})
        new = AdditiveEvents.new().combine_by_flattened_list(to_combine, 1)
        self.assertEqual(new.get_dict(), to_combine.get_dict())

    def test_AdditiveEvents_combine_by_indexed_values(self):
        to_combine = AdditiveEvents({1: 1, 2: 2})
        new = AdditiveEvents.new().combine_by_indexed_values(to_combine, 1)
        self.assertEqual(new.get_dict(), to_combine.get_dict())

    def test_AdditiveEvents_combine(self):
        to_combine = AdditiveEvents({1: 1, 2: 2})
        new = AdditiveEvents.new().combine(to_combine, 1)
        self.assertEqual(new.get_dict(), to_combine.get_dict())

    def test_AdditiveEvents_combine_works_with_low_total_occurrences_events(self):
        low_ratio_events = AdditiveEvents({1: 1, 2: 1})
        new = AdditiveEvents.new().combine(low_ratio_events, 1)
        self.assertEqual(new.get_dict(), low_ratio_events.get_dict())

    def test_AdditiveEvents_combine_works_with_high_total_occurrences_events(self):
        high_ratio_dict = {1: 10 ** 1000, 2: 10 ** 1000}
        new = AdditiveEvents.new().combine(AdditiveEvents(high_ratio_dict), 1)
        self.assertEqual(new.get_dict(), high_ratio_dict)

    def test_AdditiveEvents_one_multiple_combine_is_multiple_single_combines(self):
        to_add = AdditiveEvents({1: 2, 3: 4})
        one = AdditiveEvents.new().combine(to_add, 1)
        two = one.combine(to_add, 1)
        two_alt = AdditiveEvents.new().combine(to_add, 2)
        self.assertEqual(two.get_dict(), two_alt.get_dict())

    def test_AdditiveEvents_combine_combines_correctly(self):
        to_add = AdditiveEvents({1: 2, 2: 2})
        new = AdditiveEvents.new().combine(to_add, 2)
        self.assertEqual(new.get_dict(), {2: 4, 3: 8, 4: 4})

    def test_AdditiveEvents_combine_works_with_long_large_number_list(self):
        silly_dict = dict.fromkeys(range(-1000, 1000), 10 ** 1000)
        to_add = AdditiveEvents(silly_dict)
        new = AdditiveEvents.new().combine(to_add, 1)
        self.assertEqual(new.get_dict(), silly_dict)

    def test_AdditiveEvents_remove_removes_correctly(self):
        arbitrary_events = AdditiveEvents({-5: 3, 4: 10, 7: 1})
        five_a = AdditiveEvents.new().combine(arbitrary_events, 5)
        ten_b = AdditiveEvents.new().combine(arbitrary_events, 10)
        five_b = ten_b.remove(arbitrary_events, 5)
        self.assertEqual(five_a.get_dict(), five_b.get_dict())

    def test_AdditiveEvents_remove_works_for_large_numbers(self):
        arbitrary_large_events = AdditiveEvents({-5: 10 ** 500, 0: 5 * 10 ** 700, 3: 2 ** 1000})
        one_large = AdditiveEvents.new().combine(arbitrary_large_events, 1)
        two_large = AdditiveEvents.new().combine(arbitrary_large_events, 2)
        alt_one_large = two_large.remove(arbitrary_large_events, 1)
        self.assertEqual(one_large.get_dict(), alt_one_large.get_dict())

    def test_AdditiveEvents_combine_works_regardless_of_order(self):
        arbitrary_a = AdditiveEvents({1: 2, 3: 10 ** 456})
        arbitrary_b = AdditiveEvents({-1: 2, 0: 5})
        one_a = AdditiveEvents.new().combine(arbitrary_a, 1)
        one_a_two_b = one_a.combine(arbitrary_b, 2)
        two_b = AdditiveEvents.new().combine(arbitrary_b, 2)
        two_b_one_a = two_b.combine(arbitrary_a, 1)
        self.assertEqual(one_a_two_b.get_dict(),
                         two_b_one_a.get_dict())

    def test_AdditiveEvents_remove_removes_same_regardless_of_order(self):
        arbitrary_a = AdditiveEvents({-1: 2, 3: 5})
        arbitrary_b = AdditiveEvents({0: 9, 100: 4})
        five_a = AdditiveEvents.new().combine(arbitrary_a, 5)
        five_a_five_b = five_a.combine(arbitrary_b, 5)
        two_a_five_b = five_a_five_b.remove(arbitrary_a, 3)
        two_a_two_b = two_a_five_b.remove(arbitrary_b, 3)
        two_a_one_b = two_a_two_b.remove(arbitrary_b, 1)
        one_a_one_b = two_a_one_b.remove(arbitrary_a, 1)

        one_b = AdditiveEvents.new().combine(arbitrary_b, 1)
        one_b_one_a = one_b.combine(arbitrary_a, 1)
        self.assertEqual(one_a_one_b.get_dict(), one_b_one_a.get_dict())

    def test_AdditiveEvents_combine_defaults_to_one_time(self):
        to_use = AdditiveEvents({1: 2})
        test = AdditiveEvents.new().combine(to_use)
        self.assertEqual(test.get_dict(), {1: 2})

    def test_AdditiveEvents_combine_by_dictionary_defaults_to_one_time(self):
        to_use = AdditiveEvents({1: 2})
        test = AdditiveEvents.new().combine_by_dictionary(to_use)
        self.assertEqual(test.get_dict(), {1: 2})

    def test_AdditiveEvents_combine_by_indexed_values_defaults_to_one_time(self):
        to_use = AdditiveEvents({1: 2})
        test = AdditiveEvents.new().combine_by_indexed_values(to_use)
        self.assertEqual(test.get_dict(), {1: 2})

    def test_AdditiveEvents_combine_by_flattened_list_defaults_to_one_time(self):
        to_use = AdditiveEvents({1: 2})
        test = AdditiveEvents.new().combine_by_flattened_list(to_use)
        self.assertEqual(test.get_dict(), {1: 2})

    def test_AdditiveEvents_remove_defaults_to_one_time(self):
        to_use = AdditiveEvents({1: 2})
        test = AdditiveEvents({1: 2}).remove(to_use)
        self.assertEqual(test.get_dict(), {0: 1})



if __name__ == '__main__':
    unittest.main()
